package dynamics_of_contagion;

/**
 * A class with static methods implementing spring-mass layout
 * (the underlying idea is that every node is analogous to a mass
 * with some electric charge, and each link is a spring, if you
 * leave the system on itself, it self organizes to a structure
 * with hubs in center and hermits in the periphery)
 *
 * @author Nikolay Churkov
 *         Date: 16.12.2004
 */
public class SpringMassLayout {
    /**
     * A node with coordinates x and y
     */
    public static class Node {
        public double x;
        public double y;

        /*private Node() {
        }*/

        private Node(double x, double y) {
            this.x = x;
            this.y = y;
        }
    }

    private static class NodeEx extends Node {
        double dx;
        double dy;

        //boolean fixed;
        public NodeEx(double x, double y, int id) {
            super(x, y);
            this.id = id;
        }

        //String lbl;
        int id;
    }

    /**
     * An edge with indexes (id's) of nodes in some list
     */
    public static class Edge {
        public int from;
        public int to;
        public double length;

        /**
         * Constructs object
         * @param from - index of node
         * @param to - index of node
         * @param length - normal, relaxed length of edge
         */
        public Edge(int from, int to, double length) {
            this.from = from;
            this.to = to;
            this.length = length;
        }
    }

    private SpringMassLayout() {
    }

    /*public static Node[] arrange(int nodeCount, Edge[] edges, double x, double y, double width, double height, int nodeInitialLocationType) {
        return arrange(nodeCount, edges, x, y, width, height, 1000, nodeInitialLocationType);
    }*/

    /*public static Node[] arrange(int nodeCount, Edge[] edges, double x, double y, double width, double height, int numberOfIterations, int nodeInitialLocationType) {
        return arrange(nodeCount, edges, Math.sqrt(2.0 * width * height / (double) nodeCount), x, y, width, height, numberOfIterations, nodeInitialLocationType);
    }*/

    /**
     * Arranges nodes in graph using spring-mass layout in a rectangle
     *
     * @param nodeCount          number of nodes in graph
     * @param edges              array of edges of the graph, values "from" and "to" in edges
     *                           must be in the range [0..nodeCount-1]
     //* @param edgeLength         normal (relaxed) length of edge
     * @param x                  left coordinate of rectangle
     * @param y                  top coordinate of rectangle
     * @param width              width of rectangle
     * @param height             height of rectangle
     * @param numberOfIterations how many iterations of discretized movement should be
     *                           made before result is returned
     * @return array of nodes of the graph (with their coordinates)
     */
    public static Node[] arrange(int nodeCount, Edge[] edges, /*double edgeLength, */double x, double y, double width, double height, int numberOfIterations, int nodeInitialLocationType) {
        double xMax = x + width - 1, yMax = y + width - 1;
        double lenMax = edges[0].length * edges[0].length;
        NodeEx[] nodes = new NodeEx[nodeCount];
        NodeEx n, n1;
        int i, j;
        switch (nodeInitialLocationType) {
            case LOCATION_RING:
                double rx = width / 2, ry = height / 2,
                        cx = x + rx, cy = y + ry,
                        angleStep = 2.0 * Math.PI / nodes.length, angle;
                for (i = 0, angle = 0.0; i < nodes.length; i++, angle += angleStep) {
                    nodes[i] = new NodeEx(cx + rx * Math.cos(angle), cy + ry * Math.sin(angle), i);
                }
                break;
            default:
            case LOCATION_RANDOM:
                for (i = 0; i < nodes.length; i++) {
                    nodes[i] = new NodeEx(x + Math.random() * width, y + Math.random() * height, i);
                }
                break;
/*            case LOCATION_RECTANGLE:
                double perimeterStep = 2.0 * (width + height) / nodes.length, perimCoord = 0, perimCoordMax;
                i=0;
                perimCoordMax=width;
                while (perimCoord < perimCoordMax&&i<nodes.length) {
                    nodes[i++]=new NodeEx(x+perimCoord,y)
                    perimCoord += perimeterStep;
                }

                */
        }


        for (int iteration = 0; iteration < numberOfIterations; iteration++) {
            /******************************************************************/
            for (i = 0; i < edges.length; i++) {
                Edge e = edges[i];
                double vx = nodes[e.to].x - nodes[e.from].x;
                double vy = nodes[e.to].y - nodes[e.from].y;
                double len = Math.sqrt(vx * vx + vy * vy);
                len = (len == 0) ? .0001 : len;
                double f = (/*edgeLength*/e.length - len) / (len * 3);
                double dx = f * vx;
                double dy = f * vy;

                nodes[e.to].dx += dx;
                nodes[e.to].dy += dy;
                nodes[e.from].dx += -dx;
                nodes[e.from].dy += -dy;
            }
            //   
            for (i = 0; i < nodes.length; i++) {
                n1 = nodes[i];
                double dx = 0;
                double dy = 0;

                for (j = 0; j < nodes.length; j++) {
                    if (i == j) {
                        continue;
                    }
                    Node n2 = nodes[j];
                    double vx = n1.x - n2.x;
                    double vy = n1.y - n2.y;
                    double len = vx * vx + vy * vy;
                    if (len == 0) {
                        dx += Math.random();
                        dy += Math.random();
                    } else if (len < lenMax) {
                        dx += vx / len;
                        dy += vy / len;
                    }
                }
                double dlen = dx * dx + dy * dy;
                if (dlen > 0) {
                    dlen = Math.sqrt(dlen) / 2;
                    n1.dx += dx / dlen;
                    n1.dy += dy / dlen;
                }
            }

            for (i = 0; i < nodes.length; i++) {
                n = nodes[i];
                //if (!n.fixed) {
                n.x += Math.max(-5, Math.min(5, n.dx));
                n.y += Math.max(-5, Math.min(5, n.dy));
                //}
                if (n.x < x) {
                    n.x = x;
                } else if (n.x > x + xMax) {
                    n.x = xMax;
                }
                if (n.y < y) {
                    n.y = y;
                } else if (n.y > yMax) {
                    n.y = yMax;
                }
                // 
                n.dx /= 2;
                n.dy /= 2;
            }
            /******************************************************************/
        }
        return nodes;
    }

    /*public static void main(String[] args) {
        int curNode,nodeCnt=200;
        int linkCnt=0;
        for(int i=0;i<nodeCnt;i++) {
            curNode = i + 1+(int) (Math.random() * nodeCnt/10);
            while (curNode < nodeCnt) {
                System.out.print("," + i + "-" + curNode);
                linkCnt++;
                curNode += (int) (Math.random() * nodeCnt/10);
            }
        }
        System.out.println("\nlink count = "+linkCnt);
    }*/
    public static final int LOCATION_RANDOM = 0;
    public static final int LOCATION_RING = 1;
    //public static final int LOCATION_RECTANGLE = 2;
}
